@@ -3,7 +3,7 @@
     Author:        Jeffrey Rosenwald, Jan Wielemaker
     E-mail:        jeffrose@acm.org
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2009-2018, Jeffrey Rosenwald
+    Copyright (c)  2009-2019, Jeffrey Rosenwald
                    CWI, Amsterdam
     All rights reserved.
 
@@ -44,6 +44,10 @@
             paxos_on_change/3,                  % ?Key, ?Value, +Goal
 
             paxos_initialize/1,			% +Options
+
+            paxos_admin_key/2,                  % ?Name, ?Key
+            paxos_property/1,                   % ?Property
+            paxos_quorum_ask/4,                 % ?Templ, +Msg, -Result, +Options
                                                 % Hook support
             paxos_replicate_key/3               % +Nodes, ?Key, +Options
           ]).
@@ -107,7 +111,7 @@ Paxos is a three-phase protocol:
 
    3: Finally, the coordinator concludes the successful negotiation by
    broadcasting the agreement to the quorum in the form of a
-   paxos(changed(Key,Value) event. This is the only event that
+   paxos_changed(Key,Value) event. This is the only event that
    should be of interest to user programs.
 
 For practical reasons, we rely  on   the  partially synchronous behavior
@@ -131,7 +135,8 @@ identical for all attentive members of the quorum.|_
     paxos_on_change(?, ?, 0).
 
 :- multifile
-    paxos_message_hook/3.               % +PaxOS, +TimeOut, -Message
+    paxos_message_hook/3,               % +PaxOS, +TimeOut, -Message
+    paxos_ledger_hook/5.                % +Op, ?Key, ?Gen, ?Value, ?Status
 
 :- setting(max_sets, nonneg, 20,
            "Max Retries to get to an agreement").
@@ -144,7 +149,8 @@ identical for all attentive members of the quorum.|_
 :- setting(death_half_life, number, 10,
            "Half-time for failure score").
 :- setting(death_score, number, 100,
-           "Number of keys replicated per second").
+           "Consider a node dead if cummulative failure \c
+            score exceeds this number").
 
 
 %!  paxos_initialize(+Options) is det.
@@ -199,15 +205,15 @@ paxos_initialize :-
 %   Set administrative keys. We use a wrapper  such that we can hide the
 %   key identity.
 
-admin_key(quorum, '$paxos_quorum').
-admin_key(dead,  '$paxos_dead_nodes').
+paxos_admin_key(quorum, '$paxos_quorum').
+paxos_admin_key(dead,   '$paxos_dead_nodes').
 
 paxos_get_admin(Name, Value) :-
-    admin_key(Name, Key),
+    paxos_admin_key(Name, Key),
     paxos_get(Key, Value).
 
 paxos_set_admin(Name, Value) :-
-    admin_key(Name, Key),
+    paxos_admin_key(Name, Key),
     paxos_set(Key, Value).
 
 paxos_set_admin_bg(Name, Value) :-
@@ -352,11 +358,13 @@ set_quorum(Node, Quorum0) :-
 paxos_rejoin :-
     node(Node),
     repeat,
-        paxos_get_admin(dead, Dead0),
-        Dead is Dead0 /\ \(1<<Node),
-        (   Dead == Dead0
-        ->  true
-        ;   paxos_set_admin(dead, Dead)
+        (   paxos_get_admin(dead, Dead0)
+        ->  Dead is Dead0 /\ \(1<<Node),
+            (   Dead == Dead0
+            ->  true
+            ;   paxos_set_admin(dead, Dead)
+            )
+        ;   true
         ),
     !.
 
@@ -398,6 +406,8 @@ paxos_update_set(Set, How) :-
       (   Value == NewValue
       ->  true
       ;   paxos_set_admin(Set, NewValue)
+      ->  true
+      ;   leaving
       ),
     !.
 
@@ -490,10 +500,10 @@ life_quorum(Quorum, LifeQuorum) :-
 		 *        NETWORK STATUS	*
 		 *******************************/
 
-:- admin_key(quorum, Key),
+:- paxos_admin_key(quorum, Key),
    listen(paxos_changed(Key, Quorum),
           update_quorum(Quorum)).
-:- admin_key(dead, Key),
+:- paxos_admin_key(dead, Key),
    listen(paxos_changed(Key, Death),
           update_dead(Death)).
 
@@ -545,6 +555,23 @@ update(Clause) :-
     ;   asserta(Clause)
     ).
 
+%!  paxos_property(?Property)
+%
+%   True if Property is  a  current   property  for  the  paxos network.
+%   Defined properties are:
+%
+%     - node(?NodeID)
+%     - quorum(?NodeBitmap)
+%     - failed(?NodeBitmap)
+
+paxos_property(node(NodeID)) :-
+    node(NodeID).
+paxos_property(quorum(Quorum)) :-
+    quorum(Quorum).
+paxos_property(failed(Nodes)) :-
+    failed(Nodes).
+
+
 		 /*******************************
 		 *         INBOUND EVENTS	*
 		 *******************************/
@@ -638,6 +665,9 @@ paxos_message(claim_node(Node, Ok)) :-
     ->  Ok = false
     ;   Ok = true
     ).
+paxos_message(ask(Node, Message)) :-
+    node(Node),
+    broadcast_request(Message).
 
 
 		 /*******************************
@@ -665,7 +695,7 @@ paxos_message(claim_node(Node, Ok)) :-
 %   succeed iff the first ballot succeeds.
 %
 %   On   success,   paxos_set/1   will   also     broadcast   the   term
-%   paxos(changed(Key,Value), to the quorum.
+%   paxos_changed(Key,Value), to the quorum.
 %
 %   Options processed:
 %
@@ -697,6 +727,7 @@ paxos_set(Key, Value, Options) :-
     paxos_message(prepare(Key,Np,Rp,Value), TMO, Prepare),
     between(0, Retries, _),
       life_quorum(Quorum, Alive),
+      Alive \== 0,
       debug(paxos, 'Set: ~p -> ~p', [Key, Value]),
       collect(Quorum, false, Np, Rp, Prepare, Rps, PrepNodes),
       debug(paxos, 'Set: quorum: 0x~16r, prepared by 0x~16r, gens ~p',
@@ -737,14 +768,13 @@ intersecting(Set1, Set2) :-
 %   and NodeSet is the set (bitmask) of   Node values that replies, i.e.
 %   |NodeSet| is length(Result). The transfer stops   if  all members of
 %   the set Quorum responded or the configured timeout passed.
-%
-%   @tbd If we get a `nack` we can stop
 
 collect(Quorum, Stop, Node, Template, Message, Result, NodeSet) :-
     State = state(0),
     L0 = [dummy|_],
     Answers = list(L0),
     (   broadcast_request(Message),
+        debug(paxos(request), 'broadcast_request: ~p', [Message]),
         (   Stop
         ->  !,
             fail
@@ -766,6 +796,34 @@ collect(Quorum, Stop, Node, Template, Message, Result, NodeSet) :-
     arg(1, Answers, [_]),               % close the answer list
     L0 = [_|Result].
 
+%!  paxos_quorum_ask(?Template, +Message, -Result, +Options)
+%
+%   Ask the paxos forum for their opinion.  This predicate is not really
+%   part  of  the  paxos  protocol,  but    reuses  notably  the  quorum
+%   maintenance mechanism of this library for   asking  questions to the
+%   quorum (cluster). Message is the message to   be  asked. Result is a
+%   list of copies of Template from the quorum. Options:
+%
+%     - timeout(+Seconds)
+%       Max time to wait for a reply. Default is the setting
+%       `response_timeout`.
+%     - node(?Node)
+%       Can be used to include the replying node into Template.
+%     - quorum(?Quorum)
+%       Set/query the interviewed quorum as a bitmask
+
+paxos_quorum_ask(Template, Message, Result, Options) :-
+    option(timeout(TMO), Options, TMO),
+    option(node(Node), Options, _),
+    option(quorum(Quorum), Options, Quorum),
+    apply_default(TMO, response_timeout),
+    (   var(Quorum)
+    ->  life_quorum(Quorum, _Alive)
+    ;   true
+    ),
+    paxos_message(ask(Node, Message), TMO, BroadcastMessage),
+    collect(Quorum, false, Node, Template, BroadcastMessage, Result, _PrepNodes).
+
 %!  paxos_get(?Term) is semidet.
 %
 %   Equivalent to paxos_key(Term,Key), pasox_get(Key,Term).   I.e., Term
@@ -961,7 +1019,7 @@ replication_key(Nodes, Key) :-
 needs_replicate(Nodes, Key) :-
     ledger_current(Key, _Gen, _Value, Holders),
     Nodes /\ \Holders =\= 0,
-    \+ admin_key(_, Key).
+    \+ paxos_admin_key(_, Key).
 
 
 		 /*******************************
@@ -971,8 +1029,8 @@ needs_replicate(Nodes, Key) :-
 %!  paxos_on_change(?Term, :Goal) is det.
 %!  paxos_on_change(?Key, ?Value, :Goal) is det.
 %
-%   executes the specified Goal  when   Key  changes.  paxos_on_change/2
-%   listens for paxos(changed(Key,Value) notifications   for  Key, which
+%   Executes the specified Goal  when   Key  changes.  paxos_on_change/2
+%   listens for paxos_changed(Key,Value) notifications   for  Key, which
 %   are emitted as the result   of  successful paxos_set/3 transactions.
 %   When one is received for Key, then   Goal  is executed in a separate
 %   thread of execution.
@@ -1030,20 +1088,42 @@ key_error(E) :-
 %   @arg TimeOut is one of `-` or a time in seconds.
 
 paxos_message(Paxos:From, TMO, Message) :-
-    paxos_message_hook(paxos(Paxos):From, TMO, Message),
-    !.
+    paxos_message_raw(paxos(Paxos):From, TMO, Message).
 paxos_message(Paxos, TMO, Message) :-
-    paxos_message_hook(paxos(Paxos), TMO, Message),
+    paxos_message_raw(paxos(Paxos), TMO, Message).
+
+paxos_message_raw(Message, TMO, WireMessage) :-
+    paxos_message_hook(Message, TMO, WireMessage),
     !.
-paxos_message(Paxos, TMO, Message) :-
+paxos_message_raw(Message, TMO, WireMessage) :-
     throw(error(mode_error(det, fail,
-                           paxos:paxos_message_hook(Paxos, TMO, Message)), _)).
+                           paxos:paxos_message_hook(Message, TMO, WireMessage)), _)).
 
 
 		 /*******************************
 		 *           STORAGE		*
 		 *******************************/
 
+%!  paxos_ledger_hook(+Action, ?Key, ?Gen, ?Value, ?Holders)
+%
+%   Hook called for all operations on the ledger.  Defined actions are:
+%
+%     - current
+%       Enumerate our ledger content.
+%     - get
+%       Get a single value from our ledger.
+%     - create
+%       Create a new key in our ledger.
+%     - accept
+%       Accept a new newly proposed value for a key.  Failure causes
+%       the library to send a _NACK_ message.
+%     - set
+%       Final acceptance of Ken@Gen, providing the holders that accepted
+%       the new value.
+%     - learn
+%       Accept new keys in a new node or node that has been offline for
+%       some time.
+
 :- dynamic
     paxons_ledger/4.                    % Key, Gen, Value, Holders
 
@@ -1051,6 +1131,8 @@ paxos_message(Paxos, TMO, Message) :-
 %
 %   True when Key is a known key in my ledger.
 
+ledger_current(Key, Gen, Value, Holders) :-
+    paxos_ledger_hook(current, Key, Gen, Value, Holders).
 ledger_current(Key, Gen, Value, Holders) :-
     paxons_ledger(Key, Gen, Value, Holders),
     valid(Holders).
@@ -1062,6 +1144,11 @@ ledger_current(Key, Gen, Value, Holders) :-
 %   Note that if the value is  not   yet  acknowledged  by the leader we
 %   should not use it.
 
+ledger(Key, Gen, Value) :-
+    paxos_ledger_hook(get, Key, Gen, Value0, Holders),
+    !,
+    valid(Holders),
+    Value = Value0.
 ledger(Key, Gen, Value) :-
     paxons_ledger(Key, Gen, Value0, Holders),
     valid(Holders),
@@ -1073,6 +1160,9 @@ ledger(Key, Gen, Value) :-
 %   Create a new Key-Value pair  at   generation  Gen.  This is executed
 %   during the preparation phase.
 
+ledger_create(Key, Gen, Value) :-
+    paxos_ledger_hook(create, Key, Gen, Value, -),
+    !.
 ledger_create(Key, Gen, Value) :-
     get_time(Now),
     asserta(paxons_ledger(Key, Gen, Value, created(Now))).
@@ -1082,6 +1172,9 @@ ledger_create(Key, Gen, Value) :-
 %   Update Key to Value if the  current   generation  is older than Gen.
 %   This reflects the accept phase of the protocol.
 
+ledger_update(Key, Gen, Value) :-
+    paxos_ledger_hook(accept, Key, Gen, Value, -),
+    !.
 ledger_update(Key, Gen, Value) :-
     paxons_ledger(Key, Gen0, _Value, _Holders),
     !,
@@ -1097,6 +1190,9 @@ ledger_update(Key, Gen, Value) :-
 %
 %   The leader acknowledged that Key@Gen represents a valid new
 
+ledger_update_holders(Key, Gen, Holders) :-
+    paxos_ledger_hook(set, Key, Gen, _, Holders),
+    !.
 ledger_update_holders(Key, Gen, Holders) :-
     clause(paxons_ledger(Key, Gen, Value, Holders0), true, Ref),
     !,
@@ -1123,6 +1219,9 @@ clean_key(_, Key, Gen) :-
 %
 %   We received a learn event.
 
+ledger_learn(Key,Gen,Value) :-
+    paxos_ledger_hook(learn, Key, Gen, Value, -),
+    !.
 ledger_learn(Key,Gen,Value) :-
     paxons_ledger(Key, Gen0, Value0, _Holders),
     !,
